/*
 * tahitiwd.c
 * platform driver for WD in Tahiti platform
 *
 * Copyright (C) 2018 NetApp
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <linux/init.h>
#include <linux/io.h>
#include <linux/types.h>
#include <linux/module.h>
#include <linux/mod_devicetable.h>
#include <linux/compat.h>
#include <linux/uaccess.h>
#include <linux/platform_device.h>
#include <linux/dmi.h>
#include <linux/device.h>
#include <linux/kdev_t.h>
#include <linux/watchdog.h>
#include "tahitifpga-priv.h"

static int l1_timer = TAHITI_DEFAULT_TIMEOUT;
module_param(l1_timer, int, 0);
MODULE_PARM_DESC(l1_timer,
                 "Watchdog default L1 timeout");

struct tahiti_wdd_priv {
    struct device           *dev;
    struct watchdog_device  wdd;
};

/*
 * match tahiti platform
 */
static struct dmi_system_id platform_dmi_table[] __initdata = {
        {
                .matches = {
                        DMI_MATCH(DMI_SYS_VENDOR,"NetApp"),
                        DMI_MATCH(DMI_PRODUCT_NAME, "Tahiti"),
                }
        },
        {
                .matches = {
                        DMI_MATCH(DMI_SYS_VENDOR,"FOXCONN"),
                        DMI_MATCH(DMI_PRODUCT_NAME, "Tahiti"),
                }
        },
        {}
};
MODULE_DEVICE_TABLE(dmi, platform_dmi_table);

static bool nowayout = WATCHDOG_NOWAYOUT;
module_param(nowayout, bool, 0);
MODULE_PARM_DESC(nowayout,
		 "Watchdog cannot be stopped once started (default="
		 __MODULE_STRING(WATCHDOG_NOWAYOUT) ")");

static int tahiti_wdd_stop(struct watchdog_device *wdd)
{
	uint8_t data;

	tahiti_fpga_write(TAHITI_WD_REG, (uint8_t)~TAHITI_WD_EN);
	data = tahiti_fpga_read(TAHITI_WD_REG);
	if (data & TAHITI_WD_EN)
		return -EACCES;
	return 0;
}

static int tahiti_wdd_keepalive(struct watchdog_device *wdd)
{
	uint8_t data;
	data = tahiti_fpga_read(TAHITI_WD_REG);
	data |= TAHITI_WD_KICK;
	tahiti_fpga_write(TAHITI_WD_REG, data);
	return 0;
}

static int tahiti_wdd_start(struct watchdog_device *wdd)
{
	uint8_t data;
	tahiti_fpga_write(TAHITI_WD_REG, TAHITI_WD_EN);
	data = tahiti_fpga_read(TAHITI_WD_REG);
	if (data & TAHITI_WD_EN)
		return 0;
	else
		return -EACCES;
}

static int tahiti_wdd_set_timeout(struct watchdog_device *wdd, unsigned int timeout)
{
	wdd->timeout = timeout;
	tahiti_fpga_write(TAHITI_WD_TIMEOUT_L1, wdd->timeout);

	return 0;
}

static int tahiti_wdd_set_pretimeout(struct watchdog_device *wdd,
					unsigned int pretimeout)
{
	wdd->pretimeout = pretimeout;
	tahiti_fpga_write(TAHITI_WD_TIMEOUT_L2, wdd->pretimeout);

	return 0;
}

static int tahiti_wdd_bootstatus(void)
{
	uint8_t	data;

	data = tahiti_fpga_read(0x0c);
	if (data & 0x4)
		return WDIOF_CARDRESET;

	return 0;
}

static struct watchdog_info tahiti_wdd_info = {
	.identity       = "TAHITI Watchdog",
	.options        = WDIOF_SETTIMEOUT |
			  WDIOF_KEEPALIVEPING |
			  WDIOF_PRETIMEOUT |
			  WDIOF_MAGICCLOSE,
};

static const struct watchdog_ops tahiti_wdd_ops = {
	.owner          = THIS_MODULE,
	.start          = tahiti_wdd_start,
	.stop           = tahiti_wdd_stop,
	.ping           = tahiti_wdd_keepalive,
	.set_timeout    = tahiti_wdd_set_timeout,
	.set_pretimeout	= tahiti_wdd_set_pretimeout,
};

static int tahiti_wd_probe(struct platform_device *pdev)
{
	struct tahiti_wdd_priv *priv_data;
	struct watchdog_device *tahiti_wdd;
	int			ret;

	priv_data = devm_kzalloc(&pdev->dev, sizeof(*priv_data),
								GFP_KERNEL);
	if (!priv_data)
		return -ENOMEM;

	priv_data->dev = &pdev->dev;
	tahiti_wdd = &priv_data->wdd;

	pr_info("platform wd probe\n");
	tahiti_wdd->info = &tahiti_wdd_info;
	tahiti_wdd->ops = &tahiti_wdd_ops;
	tahiti_wdd->timeout = l1_timer;
	tahiti_fpga_write(TAHITI_WD_TIMEOUT_L1, tahiti_wdd->timeout);
	tahiti_wdd->pretimeout = TAHITI_DEFAULT_PRETIMEOUT;
	tahiti_fpga_write(TAHITI_WD_TIMEOUT_L2, tahiti_wdd->pretimeout);
	tahiti_wdd->parent = &pdev->dev;
	tahiti_wdd->bootstatus = tahiti_wdd_bootstatus();
	pr_info("nowayout = %d\n", nowayout);
	watchdog_set_nowayout(tahiti_wdd, nowayout);
	watchdog_set_drvdata(tahiti_wdd, priv_data);

	platform_set_drvdata(pdev, priv_data);

	ret = watchdog_register_device(tahiti_wdd);
	if (ret != 0)
		dev_err(&pdev->dev, "watchdog_register_device() failed: %d\n", ret);

	return ret;
}

static int tahiti_wd_remove(struct platform_device *pdev)
{
	struct tahiti_wdd_priv *priv_data = platform_get_drvdata(pdev);
	struct watchdog_device *tahiti_wdd = &priv_data->wdd;
	int ret = 0;

	if (!nowayout)
		ret = tahiti_wdd_stop(tahiti_wdd);

	watchdog_unregister_device(tahiti_wdd);
	return ret;
}

static void tahiti_wd_shutdown(struct platform_device *pdev)
{
	struct tahiti_wdd_priv *priv_data = platform_get_drvdata(pdev);
	struct watchdog_device *tahiti_wdd = &priv_data->wdd;

	if (!nowayout)
		tahiti_wdd_stop(tahiti_wdd);
}

static struct platform_driver tahiti_wd_driver = {
	.driver = {
		.name	= "tahitiwd",
		.owner	= THIS_MODULE,
	},
	.probe		= tahiti_wd_probe,
	.remove		= tahiti_wd_remove,
	.shutdown	= tahiti_wd_shutdown,
};

static int __init tahiti_wd_init(void)
{
	pr_info("calling tahiti_wd_init\n");
	if (!dmi_check_system(platform_dmi_table)) {
		pr_err("tahitiwd: unsupported system.\n");
		return -ENODEV;
        }

	platform_driver_register(&tahiti_wd_driver);
	return 0;
}

static void __exit tahiti_wd_exit(void)
{
	platform_driver_unregister(&tahiti_wd_driver);
}

module_init(tahiti_wd_init);
module_exit(tahiti_wd_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Tahiti WatchDog driver");
MODULE_AUTHOR("Hong Huang <hhuang@netapp.com>");

